from __future__ import annotations

from enum import Enum


class Status(Enum):
    """Resulting status of some action."""

    success = "success"
    failure = "failure"
    error = "error"

    def __add__(self, other: Status) -> Status:
        """Failure > Error > Success."""
        if self == Status.error:
            return Status.error

        if self == Status.failure or other == Status.failure:
            return Status.failure

        return Status.success if other == Status.success else other + self



import sys

from levo_commons.utils import syspath_prepend


def test_syspath_prepend():
    path = "./foo"
    assert path not in sys.path
    with syspath_prepend(path):
        assert sys.path[0] == path
    assert path not in sys.path


def test_syspath_prepend_nested():
    path_1 = "./foo"
    path_2 = "./bar"
    assert path_1 not in sys.path
    assert path_2 not in sys.path
    with syspath_prepend(path_1):
        assert sys.path[0] == path_1
        with syspath_prepend(path_2):
            assert sys.path[0] == path_2
            assert sys.path[1] == path_1
        assert sys.path[0] == path_1
    assert path_1 not in sys.path
    assert path_2 not in sys.path



"""Various utility functions to avoid boilerplate when performing common operations."""
import pathlib
import sys
import traceback
from contextlib import contextmanager
from typing import Generator, Union
from urllib.parse import ParseResult, urlparse

import grpc


@contextmanager
def syspath_prepend(path: Union[pathlib.Path, str]) -> Generator:
    """Temporarily prepend `path` to `sys.path` for the duration of the `with` block."""
    # NOTE. It is not thread-safe to use as is now.
    # In the future it might require an `RLock` to avoid concurrent access to `sys.path`.
    current = sys.path[:]
    # Use `insert` to put the value to the 0th position in `sys.path`. The given path will be the first one to check.
    sys.path.insert(0, str(path))
    try:
        yield
    finally:
        sys.path[:] = current


def get_grpc_channel(url: str) -> grpc.Channel:
    """Connect to a gRPC channel.

    For HTTPS it returns a secure channel and an insecure one for HTTP.
    """

    # TODO: (mike@levo.ai)
    # For whatever reason, Python or the grpc client library is no longer
    # picking up the system's trust store.  isrg_root_x1_pem is the contents
    # of...
    # https://letsencrypt.org/certs/isrgrootx1.pem
    # ...and I have verified that a file with the same contents is present in
    # /etc/ssl/certs on my system.
    #
    # Passing isrg_root_x1_pem to grpc.ssl_channel_credentials() as seen below
    # works around the TLS handshake issue we were debugging.  However, this
    # should be a SHORT TERM workaround.  We need to give some thought to how we
    # will consistently and reliably manage trust stores for any part of our
    # code that is acting as a TLS client and migrate the code below to our
    # chosen framework or convention.

    isrg_root_x1_pem = """\
-----BEGIN CERTIFICATE-----
MIIFFjCCAv6gAwIBAgIRAJErCErPDBinU/bWLiWnX1owDQYJKoZIhvcNAQELBQAw
TzELMAkGA1UEBhMCVVMxKTAnBgNVBAoTIEludGVybmV0IFNlY3VyaXR5IFJlc2Vh
cmNoIEdyb3VwMRUwEwYDVQQDEwxJU1JHIFJvb3QgWDEwHhcNMjAwOTA0MDAwMDAw
WhcNMjUwOTE1MTYwMDAwWjAyMQswCQYDVQQGEwJVUzEWMBQGA1UEChMNTGV0J3Mg
RW5jcnlwdDELMAkGA1UEAxMCUjMwggEiMA0GCSqGSIb3DQEBAQUAA4IBDwAwggEK
AoIBAQC7AhUozPaglNMPEuyNVZLD+ILxmaZ6QoinXSaqtSu5xUyxr45r+XXIo9cP
R5QUVTVXjJ6oojkZ9YI8QqlObvU7wy7bjcCwXPNZOOftz2nwWgsbvsCUJCWH+jdx
sxPnHKzhm+/b5DtFUkWWqcFTzjTIUu61ru2P3mBw4qVUq7ZtDpelQDRrK9O8Zutm
NHz6a4uPVymZ+DAXXbpyb/uBxa3Shlg9F8fnCbvxK/eG3MHacV3URuPMrSXBiLxg
Z3Vms/EY96Jc5lP/Ooi2R6X/ExjqmAl3P51T+c8B5fWmcBcUr2Ok/5mzk53cU6cG
/kiFHaFpriV1uxPMUgP17VGhi9sVAgMBAAGjggEIMIIBBDAOBgNVHQ8BAf8EBAMC
AYYwHQYDVR0lBBYwFAYIKwYBBQUHAwIGCCsGAQUFBwMBMBIGA1UdEwEB/wQIMAYB
Af8CAQAwHQYDVR0OBBYEFBQusxe3WFbLrlAJQOYfr52LFMLGMB8GA1UdIwQYMBaA
FHm0WeZ7tuXkAXOACIjIGlj26ZtuMDIGCCsGAQUFBwEBBCYwJDAiBggrBgEFBQcw
AoYWaHR0cDovL3gxLmkubGVuY3Iub3JnLzAnBgNVHR8EIDAeMBygGqAYhhZodHRw
Oi8veDEuYy5sZW5jci5vcmcvMCIGA1UdIAQbMBkwCAYGZ4EMAQIBMA0GCysGAQQB
gt8TAQEBMA0GCSqGSIb3DQEBCwUAA4ICAQCFyk5HPqP3hUSFvNVneLKYY611TR6W
PTNlclQtgaDqw+34IL9fzLdwALduO/ZelN7kIJ+m74uyA+eitRY8kc607TkC53wl
ikfmZW4/RvTZ8M6UK+5UzhK8jCdLuMGYL6KvzXGRSgi3yLgjewQtCPkIVz6D2QQz
CkcheAmCJ8MqyJu5zlzyZMjAvnnAT45tRAxekrsu94sQ4egdRCnbWSDtY7kh+BIm
lJNXoB1lBMEKIq4QDUOXoRgffuDghje1WrG9ML+Hbisq/yFOGwXD9RiX8F6sw6W4
avAuvDszue5L3sz85K+EC4Y/wFVDNvZo4TYXao6Z0f+lQKc0t8DQYzk1OXVu8rp2
yJMC6alLbBfODALZvYH7n7do1AZls4I9d1P4jnkDrQoxB3UqQ9hVl3LEKQ73xF1O
yK5GhDDX8oVfGKF5u+decIsH4YaTw7mP3GFxJSqv3+0lUFJoi5Lc5da149p90Ids
hCExroL1+7mryIkXPeFM5TgO9r0rvZaBFOvV2z0gp35Z0+L4WPlbuEjN/lxPFin+
HlUjr8gRsI3qfJOQFy/9rKIJR0Y/8Omwt/8oTWgy1mdeHmmjk7j1nYsvC9JSQ6Zv
MldlTTKB3zhThV1+XWYp6rjd5JW1zbVWEkLNxE7GJThEUG3szgBVGP7pSWTUTsqX
nLRbwHOoq7hHwg==
-----END CERTIFICATE-----""".encode(
        "ascii"
    )
    parsed = urlparse(url)
    address = get_grpc_address(parsed)
    if parsed.scheme == "https":
        return grpc.secure_channel(
            address, grpc.ssl_channel_credentials(root_certificates=isrg_root_x1_pem)
        )
    return grpc.insecure_channel(address)


def get_grpc_address(parsed: ParseResult) -> str:
    if parsed.port is not None:
        port = parsed.port
    else:
        port = 443 if parsed.scheme == "https" else 80
    return f"{parsed.hostname}:{port}"


def format_exception(error: Exception, include_traceback: bool = False) -> str:
    """Format exception as text."""
    error_type = type(error)
    if include_traceback:
        lines = traceback.format_exception(error_type, error, error.__traceback__)
    else:
        lines = traceback.format_exception_only(error_type, error)
    return "".join(lines)



from pathlib import Path
from typing import Any, Callable, Dict, List, Set, Tuple, Union

PathLike = Union[Path, str]

Query = Dict[str, Any]
# Body can be of any Python type that corresponds to JSON Schema types + `bytes`
Body = Union[List, Dict[str, Any], str, int, float, bool, bytes]
PathParameters = Dict[str, Any]
Headers = Dict[str, Any]
Cookies = Dict[str, Any]
FormData = Dict[str, Any]


class NotSet:
    pass


# A filter for path / method
Filter = Union[str, List[str], Tuple[str], Set[str], NotSet]

RawAuth = Tuple[str, str]
# Generic test with any arguments and no return
GenericTest = Callable[..., None]



from levo_commons.config import AuthConfig, PlanConfig


def test_config():
    config = PlanConfig(
        target_url="http://localhost",
        headers={"myheader": "myvalue"},
        auth_config=AuthConfig(auth_type="basic", username="test", password="test"),
    )
    assert len(config.headers) == 1
    assert config.asdict() == {
        "target_url": "http://localhost",
        "spec_path": None,
        "test_plan_path": None,
        "auth": None,
        "auth_type": None,
        "report_to_saas": True,
        "headers": {"myheader": "myvalue"},
        "auth_config": {
            "auth_type": "basic",
            "username": "test",
            "password": "test",
            "api_key": None,
            "token": None,
        },
    }



from typing import Any, Dict, List, Union

import attr


@attr.s(slots=True, repr=False)
class FailureContext:
    """Additional data specific to certain failure kind."""

    # Short description of what happened
    title: str
    # A longer one
    message: str


@attr.s(slots=True, repr=False)
class ValidationErrorContext(FailureContext):
    """Additional information about JSON Schema validation errors."""

    title = "Non-conforming response payload"
    message = "Response does not conform to the defined schema"
    validation_message: str = attr.ib()
    schema_path: List[Union[str, int]] = attr.ib()
    schema: Union[Dict[str, Any], bool] = attr.ib()
    instance_path: List[Union[str, int]] = attr.ib()
    instance: Union[None, bool, float, str, list, Dict[str, Any]] = attr.ib()


@attr.s(slots=True, repr=False)
class JSONDecodeErrorContext(FailureContext):
    """Failed to decode JSON."""

    title = "JSON deserialization error"
    message = "Response is not a valid JSON"
    validation_message: str = attr.ib()
    document: str = attr.ib()
    position: int = attr.ib()
    lineno: int = attr.ib()
    colno: int = attr.ib()


@attr.s(slots=True, repr=False)
class ServerError(FailureContext):
    title = "Internal server error"
    message = "Server got itself in trouble"
    status_code: int = attr.ib()


@attr.s(slots=True, repr=False)
class MissingContentType(FailureContext):
    """Content type header is missing."""

    title = "Missing Content-Type header"
    message = "Response is missing the `Content-Type` header"
    media_types: List[str] = attr.ib()


@attr.s(slots=True, repr=False)
class UndefinedContentType(FailureContext):
    """Response has Content-Type that is not defined in the schema."""

    title = "Undefined Content-Type"
    message = "Response has `Content-Type` that is not declared in the schema"
    content_type: str = attr.ib()
    defined_content_types: List[str] = attr.ib()


@attr.s(slots=True, repr=False)
class UndefinedStatusCode(FailureContext):
    """Response has a status code that is not defined in the schema."""

    title = "Undefined status code"
    message = "Response has a status code that is not declared in the schema"
    # Response's status code
    status_code: int = attr.ib()
    # Status codes as defined in schema
    defined_status_codes: List[str] = attr.ib()
    # Defined status code with expanded wildcards
    allowed_status_codes: List[int] = attr.ib()


@attr.s(slots=True, repr=False)
class MissingHeaders(FailureContext):
    """Some required headers are missing."""

    title = "Missing required headers"
    message = "Response is missing headers required by the schema"
    missing_headers: List[str] = attr.ib()


@attr.s(slots=True, repr=False)
class MalformedMediaType(FailureContext):
    """Media type name is malformed.

    Example: `application-json` instead of `application/json`
    """

    actual: str = attr.ib()
    defined: str = attr.ib()
    title: str = attr.ib("Malformed media type name")
    message: str = attr.ib("Media type name is not valid")


@attr.s(slots=True, repr=False)
class ResponseTimeExceeded(FailureContext):
    """Response took longer than expected."""

    title = "Response time exceeded"
    message = "Response time exceeds the deadline"
    elapsed: float = attr.ib()
    deadline: int = attr.ib()


@attr.s(slots=True, repr=False)
class RequestTimeout(FailureContext):
    """Request took longer than timeout."""

    title = "Request timeout"
    message = "The request timed out"
    timeout: int = attr.ib()



from __future__ import annotations

import base64
import datetime
import logging
import threading
from logging import LogRecord
from typing import Any, Dict, List, Optional, cast

import attr
import curlify
import requests

from .events import Payload
from .failures import FailureContext
from .status import Status
from .types import Headers
from .utils import format_exception


@attr.s(slots=True, repr=False)
class Case:
    text_lines: Optional[List[str]] = attr.ib(default=None)
    requests_code: Optional[str] = attr.ib(default=None)
    curl_code: Optional[str] = attr.ib(default=None)


@attr.s(slots=True, repr=False)
class Check:
    """Single check run result."""

    name: str = attr.ib()
    value: Status = attr.ib()
    response: Optional[requests.Response] = attr.ib()
    elapsed: float = attr.ib()
    example: Optional[Case] = attr.ib(default=None)
    message: Optional[str] = attr.ib(default=None)
    # Failure-specific context
    context: Optional[FailureContext] = attr.ib(default=None)
    request: Optional[requests.PreparedRequest] = attr.ib(default=None)


@attr.s(slots=True, repr=False)
class Request:
    """Request data extracted from TestCase."""

    method: str = attr.ib()
    uri: str = attr.ib()
    body: Optional[str] = attr.ib()
    headers: Headers = attr.ib()

    @classmethod
    def from_prepared_request(cls, prepared: requests.PreparedRequest) -> "Request":
        """A prepared request version is already stored in `requests.Response`."""
        body = prepared.body

        if isinstance(body, str):
            # can be a string for `application/x-www-form-urlencoded`
            body = body.encode("utf-8")

        # these values have `str` type at this point
        uri = cast(str, prepared.url)
        method = cast(str, prepared.method)
        return cls(
            uri=uri,
            method=method,
            headers={key: [value] for (key, value) in prepared.headers.items()},
            body=base64.b64encode(body).decode() if body is not None else body,
        )

    def as_curl_command(self) -> str:
        """Construct a curl command for a given Request."""
        prepared_request = requests.PreparedRequest()
        prepared_request.prepare(
            url=self.uri,
            method=self.method,
            headers=self.headers,
            data=base64.b64decode(self.body.encode()).decode()
            if self.body is not None
            else self.body,
        )

        return curlify.to_curl(prepared_request)


def serialize_payload(payload: bytes) -> str:
    return base64.b64encode(payload).decode()


@attr.s(slots=True, repr=False)
class Response:
    """Unified response data."""

    status_code: int = attr.ib()
    message: str = attr.ib()
    headers: Dict[str, List[str]] = attr.ib()
    body: Optional[str] = attr.ib()
    encoding: Optional[str] = attr.ib()
    http_version: str = attr.ib()
    elapsed: float = attr.ib()

    @classmethod
    def from_requests(cls, response: requests.Response) -> "Response":
        """Create a response from requests.Response."""
        headers = {
            name: response.raw.headers.getlist(name)
            for name in response.raw.headers.keys()
        }
        # Similar to http.client:319 (HTTP version detection in stdlib's `http` package)
        http_version = "1.0" if response.raw.version == 10 else "1.1"

        def is_empty(_response: requests.Response) -> bool:
            # Assume the response is empty if:
            #   - no `Content-Length` header
            #   - no chunks when iterating over its content
            return (
                "Content-Length" not in headers and list(_response.iter_content()) == []
            )

        body = None if is_empty(response) else serialize_payload(response.content)
        return cls(
            status_code=response.status_code,
            message=response.reason,
            body=body,
            encoding=response.encoding,
            headers=headers,
            http_version=http_version,
            elapsed=response.elapsed.total_seconds(),
        )


@attr.s(slots=True)
class Interaction:
    """A single interaction with the target app."""

    request: Request = attr.ib()
    response: Response = attr.ib()
    checks: List[Check] = attr.ib()
    status: Status = attr.ib()
    recorded_at: str = attr.ib(factory=lambda: datetime.datetime.now().isoformat())

    @classmethod
    def from_requests(
        cls, response: requests.Response, status: Status, checks: List[Check]
    ) -> "Interaction":
        return cls(
            request=Request.from_prepared_request(response.request),
            response=Response.from_requests(response),
            status=status,
            checks=checks,
        )


@attr.s(slots=True, repr=False)
class TestResult:
    """Result of a single test."""

    __test__ = False
    checks: List[Check] = attr.ib(factory=list)
    errors: List[Exception] = attr.ib(factory=list)
    interactions: List[Interaction] = attr.ib(factory=list)
    logs: List[LogRecord] = attr.ib(factory=list)
    is_errored: bool = attr.ib(default=False)
    summary: Optional[str] = attr.ib(default=None)
    # To show a proper reproduction code if an error happens and there is no way to get actual headers that were
    # sent over the network. Or there could be no actual requests at all
    overridden_headers: Optional[Dict[str, Any]] = attr.ib(default=None)

    def __add__(self, other: TestResult) -> TestResult:
        return TestResult(
            checks=self.checks + other.checks,
            errors=self.errors + other.errors,
            interactions=self.interactions + other.interactions,
            logs=self.logs + other.logs,
            is_errored=self.is_errored or other.is_errored,
            summary=self.summary,
        )

    def mark_errored(self) -> None:
        self.is_errored = True

    @property
    def has_errors(self) -> bool:
        return bool(self.errors)

    @property
    def has_failures(self) -> bool:
        return any(check.value == Status.failure for check in self.checks)

    @property
    def has_logs(self) -> bool:
        return bool(self.logs)

    def add_success(
        self, name: str, response: requests.Response, elapsed: float
    ) -> Check:
        check = Check(
            name=name,
            value=Status.success,
            response=response,
            elapsed=elapsed,
            request=None,
        )
        self.checks.append(check)
        return check

    def add_failure(
        self,
        name: str,
        response: Optional[requests.Response],
        elapsed: float,
        message: str,
        context: Optional[FailureContext],
        request: Optional[requests.PreparedRequest] = None,
    ) -> Check:
        check = Check(
            name=name,
            value=Status.failure,
            response=response,
            elapsed=elapsed,
            message=message,
            context=context,
            request=request,
        )
        self.checks.append(check)
        return check

    def add_error(self, exception: Exception) -> None:
        self.errors.append(exception)

    def store_requests_response(
        self, response: requests.Response, status: Status, checks: List[Check]
    ) -> None:
        self.interactions.append(Interaction.from_requests(response, status, checks))


@attr.s(slots=True)
class SerializedError:
    exception: str = attr.ib()
    exception_with_traceback: str = attr.ib()
    title: Optional[str] = attr.ib()

    @classmethod
    def from_error(
        cls,
        exception: Exception,
        title: Optional[str] = None,
    ) -> "SerializedError":
        return cls(
            exception=format_exception(exception),
            exception_with_traceback=format_exception(exception, True),
            title=title,
        )


@attr.s(slots=True)
class SerializedInteraction:
    request: Request = attr.ib()
    response: Response = attr.ib()
    status: Status = attr.ib()
    recorded_at: str = attr.ib()

    @classmethod
    def from_interaction(cls, interaction: Interaction) -> "SerializedInteraction":
        return cls(
            request=interaction.request,
            response=interaction.response,
            status=interaction.status,
            recorded_at=interaction.recorded_at,
        )


@attr.s(slots=True)
class SerializedTestResult:
    has_failures: bool = attr.ib()
    has_errors: bool = attr.ib()
    has_logs: bool = attr.ib()
    is_errored: bool = attr.ib()
    logs: List[str] = attr.ib()
    errors: List[SerializedError] = attr.ib()
    interactions: List[SerializedInteraction] = attr.ib()

    @classmethod
    def from_test_result(cls, result: TestResult) -> "SerializedTestResult":
        formatter = logging.Formatter(
            "[%(asctime)s] %(levelname)s in %(module)s: %(message)s"
        )
        return cls(
            has_failures=result.has_failures,
            has_errors=result.has_errors,
            has_logs=result.has_logs,
            is_errored=result.is_errored,
            logs=[formatter.format(record) for record in result.logs],
            errors=[SerializedError.from_error(error) for error in result.errors],
            interactions=[
                SerializedInteraction.from_interaction(interaction)
                for interaction in result.interactions
            ],
        )


@attr.s(slots=True)
class InitializedPayload(Payload):
    plan_id: str = attr.ib()
    plan_name: str = attr.ib()
    workspace_id: str = attr.ib()

    # The target URL against which the tests are running
    target_url: str = attr.ib()

    # A dictionary to hold some generic attributes like entity metadata, etc.
    attributes: dict = attr.ib(factory=dict)


@attr.s(slots=True)
class BeforeTestSuiteExecutionPayload(Payload):
    name: str = attr.ib()
    test_suite_id: Optional[str] = attr.ib(default=None)
    modules: List[str] = attr.ib(factory=list)


@attr.s(slots=True)
class AfterTestSuiteExecutionPayload(Payload):
    name: str = attr.ib()
    test_suite_id: Optional[str] = attr.ib(default=None)
    errored: bool = attr.ib(default=False)
    thread_id: int = attr.ib(factory=threading.get_ident)


@attr.s(slots=True)
class BeforeTestExecutionPayload(Payload):
    test_case_id: str = attr.ib()
    name: str = attr.ib()
    method: str = attr.ib()
    path: str = attr.ib()
    relative_path: str = attr.ib()
    test_suite_id: Optional[str] = attr.ib(default=None)
    description: Optional[str] = attr.ib(default=None)

    # The current level of recursion during stateful testing
    recursion_level: int = attr.ib(default=0)

    module: Optional[str] = attr.ib(default=None)


@attr.s(slots=True)
class AfterTestExecutionPayload(Payload):
    test_case_id: str = attr.ib()
    name: str = attr.ib()
    method: str = attr.ib()
    path: str = attr.ib()
    relative_path: str = attr.ib()
    status: Status = attr.ib()
    elapsed_time: float = attr.ib()
    result: Optional[TestResult] = attr.ib()
    test_suite_id: Optional[str] = attr.ib(default=None)
    thread_id: int = attr.ib(factory=threading.get_ident)

    def __add__(self, other: AfterTestExecutionPayload) -> AfterTestExecutionPayload:
        result = None
        if not self.result:
            result = other.result
        elif not other.result:
            result = self.result
        else:
            result = self.result + other.result

        return AfterTestExecutionPayload(
            test_case_id=self.test_case_id,
            name=self.name,
            method=self.method,
            path=self.path,
            relative_path=self.path,
            status=self.status + other.status,
            elapsed_time=self.elapsed_time + other.elapsed_time,
            result=result,
            test_suite_id=self.test_suite_id,
            thread_id=self.thread_id,
        )


@attr.s(slots=True)
class AfterStepExecutionPayload(Payload):
    test_case_id: str = attr.ib()
    name: str = attr.ib()
    sequence: int = attr.ib()
    elapsed_time: float = attr.ib()
    interaction: Interaction = attr.ib()


@attr.s(slots=True)
class FinishedPayload(Payload):
    plan_id: str = attr.ib()
    has_failures: bool = attr.ib(default=False)
    has_errors: bool = attr.ib(default=False)
    generic_errors: List[Exception] = attr.ib(factory=list)



#!/usr/bin/env python

import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()


install_requires = [
    "attrs>=20.3",
    "curlify>=2.2.1",
    "grpcio>=1.37.0,<2.0.0",
    "requests>=2.25.1",
]

setuptools.setup(
    name="levo_commons",
    version="0.1.13",
    author="Levo Inc",
    author_email="info@levo.ai",
    description="Common code between Levo CLI and test plans.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/levoai/levo-commons",
    project_urls={
        "Bug Tracker": "https://github.com/levo/levo-commons/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    install_requires=install_requires,
    extras_require={"test": ["pytest>=6.0"]},
    python_requires=">=3.7",
)
